@@ -5,15 +5,17 @@ import collections
 from functools import lru_cache
 from copy import copy, deepcopy
 from collections import defaultdict, OrderedDict
-from typing import Mapping, Tuple, Type
+from typing import Mapping, Tuple, Type, Union, Optional, List, Any
+
+from linkml_runtime.linkml_model import PermissibleValue, PermissibleValueText
 from linkml_runtime.utils.namespaces import Namespaces
 from deprecated.classic import deprecated
 from linkml_runtime.utils.context_utils import parse_import_map, map_import
 from linkml_runtime.utils.pattern import PatternResolver
 from linkml_runtime.linkml_model.meta import *
 from enum import Enum
-logger = logging.getLogger(__name__)
 
+logger = logging.getLogger(__name__)
 
 MAPPING_TYPE = str  ## e.g. broad, exact, related, ...
 CACHE_SIZE = 1024
@@ -52,10 +54,10 @@ def _closure(f, x, reflexive=True, depth_first=True, **kwargs):
             todo = todo[1:]
         visited.append(i)
         vals = f(i)
-        for v in vals:
-            if v not in visited:
-                todo.append(v)
-                if v not in rv:
+        if vals is not None:
+            for v in vals:
+                if v not in visited and v not in rv:
+                    todo.append(v)
                     rv.append(v)
     return rv
 
@@ -105,20 +107,23 @@ class SchemaView(object):
     schema: SchemaDefinition = None
     schema_map: Dict[SchemaDefinitionName, SchemaDefinition] = None
     importmap: Optional[Mapping[str, str]] = None
+    """Optional mapping between schema names and local paths/URLs"""
     modifications: int = 0
     uuid: str = None
 
     def __init__(self, schema: Union[str, SchemaDefinition],
-                 importmap: Optional[Mapping[str, str]] = None):
+                 importmap: Optional[Mapping[str, str]] = None, merge_imports: bool = False):
         if isinstance(schema, str):
             schema = load_schema_wrap(schema)
         self.schema = schema
         self.schema_map = {schema.name: schema}
         self.importmap = parse_import_map(importmap, self.base_dir) if self.importmap is not None else dict()
+        if merge_imports:
+            self.merge_imports()
         self.uuid = str(uuid.uuid4())
 
     def __key(self):
-        return (self.schema.id, self.uuid, self.modifications)
+        return self.schema.id, self.uuid, self.modifications
 
     def __eq__(self, other):
         if isinstance(other, SchemaView):
@@ -144,17 +149,20 @@ class SchemaView(object):
         sname = map_import(self.importmap, self.namespaces, imp)
         logging.info(f'Loading schema {sname} from {from_schema.source_file}')
         schema = load_schema_wrap(sname + '.yaml',
-                                  base_dir=os.path.dirname(from_schema.source_file) if from_schema.source_file else None)
+                                  base_dir=os.path.dirname(
+                                      from_schema.source_file) if from_schema.source_file else None)
         return schema
 
     @lru_cache()
-    def imports_closure(self, traverse=True, inject_metadata=True) -> List[SchemaDefinitionName]:
+    def imports_closure(self, imports: bool = True, traverse=True, inject_metadata=True) -> List[SchemaDefinitionName]:
         """
         Return all imports
 
         :param traverse: if true, traverse recursively
         :return: all schema names in the transitive reflexive imports closure
         """
+        if not imports:
+            return [self.schema.name]
         if self.schema_map is None:
             self.schema_map = {self.schema.name: self.schema}
         closure = []
@@ -183,9 +191,8 @@ class SchemaView(object):
                         a.from_schema = s.id
         return closure
 
-
     @lru_cache()
-    def all_schema(self, imports: True) -> List[SchemaDefinition]:
+    def all_schema(self, imports: bool = True) -> List[SchemaDefinition]:
         """
         :param imports: include imports closure
         :return: all schemas
@@ -267,7 +274,8 @@ class SchemaView(object):
         return self.all_slots(**kwargs)
 
     @lru_cache()
-    def all_slots(self, ordered_by=OrderedBy.PRESERVE, imports=True, attributes=True) -> Dict[SlotDefinitionName, SlotDefinition]:
+    def all_slots(self, ordered_by=OrderedBy.PRESERVE, imports=True, attributes=True) -> Dict[
+        SlotDefinitionName, SlotDefinition]:
         """
         :param ordered_by: an enumerated parameter that returns all the slots in the order specified.
         :param imports: include imports closure
@@ -308,7 +316,6 @@ class SchemaView(object):
         """
         return self._get_dict(ENUMS, imports)
 
-
     @deprecated("Use `all_types` instead")
     @lru_cache()
     def all_type(self, imports=True) -> Dict[TypeDefinitionName, TypeDefinition]:
@@ -378,7 +385,7 @@ class SchemaView(object):
         for s in schemas:
             # get the value of element name from the schema, if empty, return empty dictionary.
             d1 = getattr(s, slot_name, {})
-            # {**d,**d1} syntax merges dictionary a and b into a single dictionary, removing duplicates.
+            # {**d,**d1} syntax merges dictionary d and d1 into a single dictionary, removing duplicates.
             d = {**d, **d1}
 
         return d
@@ -411,7 +418,6 @@ class SchemaView(object):
             m[camelcase(s.name)] = s
         return m
 
-
     @lru_cache()
     def in_schema(self, element_name: ElementName) -> SchemaDefinitionName:
         """
@@ -532,14 +538,32 @@ class SchemaView(object):
         return self._parents(cls, imports, mixins, is_a)
 
     @lru_cache()
-    def enum_parents(self, enum_name: ENUM_NAME, imports=True, mixins=True, is_a=True) -> List[EnumDefinitionName]:
+    def enum_parents(self, enum_name: ENUM_NAME, imports=False, mixins=False, is_a=True) -> List[EnumDefinitionName]:
         """
         :param enum_name: child enum name
-        :param imports: include import closure
-        :param mixins: include mixins (default is True)
+        :param imports: include import closure (False)
+        :param mixins: include mixins (default is False)
         :return: all direct parent enum names (is_a and mixins)
         """
-        return []
+        e = self.get_enum(enum_name, strict=True)
+        return self._parents(e, imports, mixins, is_a=is_a)
+
+    @lru_cache()
+    def permissible_value_parent(self, permissible_value: str, enum_name: ENUM_NAME) -> Union[
+        str, PermissibleValueText, None, ValueError]:
+        """
+        :param enum_name: child enum name
+        :param permissible_value: permissible value
+        :return: all direct parent enum names (is_a)
+        """
+        enum = self.get_enum(enum_name, strict=True)
+        if enum:
+            if permissible_value in enum.permissible_values:
+                pv = enum.permissible_values[permissible_value]
+                if pv.is_a:
+                    return [pv.is_a]
+        else:
+            return []
 
     @lru_cache()
     def slot_parents(self, slot_name: SLOT_NAME, imports=True, mixins=True, is_a=True) -> List[SlotDefinitionName]:
@@ -608,11 +632,26 @@ class SchemaView(object):
         """
         return _closure(lambda x: self.class_parents(x, imports=imports, mixins=mixins, is_a=is_a),
                         class_name,
-                        reflexive=reflexive,  depth_first=depth_first)
+                        reflexive=reflexive, depth_first=depth_first)
+
+    @lru_cache()
+    def permissible_value_ancestors(self, permissible_value_text: str,
+                                    enum_name: ENUM_NAME,
+                                    reflexive=True,
+                                    depth_first=True) -> List[str]:
+        """
+        Closure of permissible_value_parents method
+        :enum
+        """
+
+        return _closure(lambda x: self.permissible_value_parent(x, enum_name),
+                        permissible_value_text,
+                        reflexive=reflexive,
+                        depth_first=depth_first)
 
     @lru_cache()
     def enum_ancestors(self, enum_name: ENUM_NAME, imports=True, mixins=True, reflexive=True, is_a=True,
-                        depth_first=True) -> List[EnumDefinitionName]:
+                       depth_first=True) -> List[EnumDefinitionName]:
         """
         Closure of enum_parents method
 
@@ -626,10 +665,11 @@ class SchemaView(object):
         """
         return _closure(lambda x: self.enum_parents(x, imports=imports, mixins=mixins, is_a=is_a),
                         enum_name,
-                        reflexive=reflexive,  depth_first=depth_first)
+                        reflexive=reflexive, depth_first=depth_first)
 
     @lru_cache()
-    def type_ancestors(self, type_name: TYPES, imports=True, reflexive=True, depth_first=True) -> List[TypeDefinitionName]:
+    def type_ancestors(self, type_name: TYPES, imports=True, reflexive=True, depth_first=True) -> List[
+        TypeDefinitionName]:
         """
         All ancestors of a type via typeof
 
@@ -641,10 +681,11 @@ class SchemaView(object):
         """
         return _closure(lambda x: self.type_parents(x, imports=imports),
                         type_name,
-                        reflexive=reflexive,  depth_first=depth_first)
+                        reflexive=reflexive, depth_first=depth_first)
 
     @lru_cache()
-    def slot_ancestors(self, slot_name: SLOT_NAME, imports=True, mixins=True, reflexive=True, is_a=True) -> List[SlotDefinitionName]:
+    def slot_ancestors(self, slot_name: SLOT_NAME, imports=True, mixins=True, reflexive=True, is_a=True) -> List[
+        SlotDefinitionName]:
         """
         Closure of slot_parents method
 
@@ -660,7 +701,8 @@ class SchemaView(object):
                         reflexive=reflexive)
 
     @lru_cache()
-    def class_descendants(self, class_name: CLASS_NAME, imports=True, mixins=True, reflexive=True, is_a=True) -> List[ClassDefinitionName]:
+    def class_descendants(self, class_name: CLASS_NAME, imports=True, mixins=True, reflexive=True, is_a=True) -> List[
+        ClassDefinitionName]:
         """
         Closure of class_children method
 
@@ -671,10 +713,12 @@ class SchemaView(object):
         :param reflexive: include self in set of descendants
         :return: descendants class names
         """
-        return _closure(lambda x: self.class_children(x, imports=imports, mixins=mixins, is_a=is_a), class_name, reflexive=reflexive)
+        return _closure(lambda x: self.class_children(x, imports=imports, mixins=mixins, is_a=is_a), class_name,
+                        reflexive=reflexive)
 
     @lru_cache()
-    def slot_descendants(self, slot_name: SLOT_NAME, imports=True, mixins=True, reflexive=True, is_a=True) -> List[SlotDefinitionName]:
+    def slot_descendants(self, slot_name: SLOT_NAME, imports=True, mixins=True, reflexive=True, is_a=True) -> List[
+        SlotDefinitionName]:
         """
         Closure of slot_children method
 
@@ -685,7 +729,8 @@ class SchemaView(object):
         :param reflexive: include self in set of descendants
         :return: descendants slot names
         """
-        return _closure(lambda x: self.slot_children(x, imports=imports, mixins=mixins, is_a=is_a), slot_name, reflexive=reflexive)
+        return _closure(lambda x: self.slot_children(x, imports=imports, mixins=mixins, is_a=is_a), slot_name,
+                        reflexive=reflexive)
 
     @lru_cache()
     def class_roots(self, imports=True, mixins=True, is_a=True) -> List[ClassDefinitionName]:
@@ -713,7 +758,6 @@ class SchemaView(object):
                 for c in self.all_classes(imports=imports)
                 if self.class_children(c, mixins=mixins, is_a=is_a, imports=imports) == []]
 
-
     @lru_cache()
     def slot_roots(self, imports=True, mixins=True) -> List[SlotDefinitionName]:
         """
@@ -876,7 +920,8 @@ class SchemaView(object):
         return applicable_elements
 
     @lru_cache()
-    def get_mappings(self, element_name: ElementName = None, imports=True, expand=False) -> Dict[MAPPING_TYPE, List[URIorCURIE]]:
+    def get_mappings(self, element_name: ElementName = None, imports=True, expand=False) -> Dict[
+        MAPPING_TYPE, List[URIorCURIE]]:
         """
         Get all mappings for a given element
 
@@ -998,9 +1043,9 @@ class SchemaView(object):
         e = self.get_element(element_name, imports=imports)
         return {k: v.value for k, v in e.annotations.items()}
 
-
     @lru_cache()
-    def class_slots(self, class_name: CLASS_NAME, imports=True, direct=False, attributes=True) -> List[SlotDefinitionName]:
+    def class_slots(self, class_name: CLASS_NAME, imports=True, direct=False, attributes=True) -> List[
+        SlotDefinitionName]:
         """
         :param class_name:
         :param imports: include imports closure
@@ -1025,7 +1070,8 @@ class SchemaView(object):
         return slots_nr
 
     @lru_cache()
-    def induced_slot(self, slot_name: SLOT_NAME, class_name: CLASS_NAME = None, imports=True, mangle_name=False) -> SlotDefinition:
+    def induced_slot(self, slot_name: SLOT_NAME, class_name: CLASS_NAME = None, imports=True,
+                     mangle_name=False) -> SlotDefinition:
         """
         Given a slot, in the context of a particular class, yield a dynamic SlotDefinition that
         has all properties materialized.
@@ -1045,33 +1091,31 @@ class SchemaView(object):
             cls = None
         # attributes take priority over schema-level slot definitions, IF
         # the attributes is declared for the class or an ancestor
+        slot_comes_from_attribute = False
         if cls:
+            # traverse ancestors (reflexive), starting with
+            # the main class
             for an in self.class_ancestors(class_name):
                 a = self.get_class(an, imports)
                 if slot_name in a.attributes:
                     slot = a.attributes[slot_name]
+                    slot_comes_from_attribute = True
                     break
-        islot = None
-        if slot is not None:
-            # case 1: there is an explicit declaration of the slot
-            islot = deepcopy(slot)
+        else:
+            slot = self.get_slot(slot_name, imports, attributes=True)
+        if slot is None:
+            raise ValueError(f"Slot {slot_name} as an attribute of {class_name} ancestors"
+                             "or as a slot definition in the schema")
+        # copy the slot, as it will be modified
+        induced_slot = deepcopy(slot)
+        if not slot_comes_from_attribute:
             slot_anc_names = self.slot_ancestors(slot_name, reflexive=True)
             # inheritable slot: first propagate from ancestors
             for anc_sn in reversed(slot_anc_names):
                 anc_slot = self.get_slot(anc_sn, attributes=False)
                 for metaslot_name in SlotDefinition._inherited_slots:
                     if getattr(anc_slot, metaslot_name, None):
-                        setattr(islot, metaslot_name, deepcopy(getattr(anc_slot, metaslot_name)))
-            # then override with this
-            #for metaslot_name in SlotDefinition._inherited_slots:
-            #    if getattr(slot, metaslot_name, None):
-            #        setattr(islot, metaslot_name, deepcopy(getattr(slot, metaslot_name)))
-        else:
-            slot = self.get_slot(slot_name, imports, attributes=True)
-            if slot is None:
-                raise ValueError(f'No such slot: {slot_name} and no attribute by that name in ancestors of {class_name}')
-            islot = deepcopy(slot)
-
+                        setattr(induced_slot, metaslot_name, deepcopy(getattr(anc_slot, metaslot_name)))
         COMBINE = {
             'maximum_value': lambda x, y: min(x, y),
             'minimum_value': lambda x, y: max(x, y),
@@ -1080,15 +1124,13 @@ class SchemaView(object):
         for metaslot_name in self._metaslots_for_slot():
             # inheritance of slots; priority order
             #   slot-level assignment < ancestor slot_usage < self slot_usage
-            v = getattr(islot, metaslot_name, None)
+            v = getattr(induced_slot, metaslot_name, None)
             if not cls:
                 propagated_from = []
-            elif metaslot_name in SlotDefinition._inherited_slots:
-                propagated_from = self.class_ancestors(class_name, reflexive=True, mixins=True)
             else:
-                propagated_from = [class_name]
+                propagated_from = self.class_ancestors(class_name, reflexive=True, mixins=True)
             for an in reversed(propagated_from):
-                islot.owner = an
+                induced_slot.owner = an
                 a = self.get_class(an, imports)
                 anc_slot_usage = a.slot_usage.get(slot_name, {})
                 v2 = getattr(anc_slot_usage, metaslot_name, None)
@@ -1101,23 +1143,23 @@ class SchemaView(object):
                     else:
                         if v2 is not None:
                             v = v2
-                            logging.debug(f'{v} takes precedence over {v2} for {islot.name}.{metaslot_name}')
+                            logging.debug(f'{v} takes precedence over {v2} for {induced_slot.name}.{metaslot_name}')
             if v is None:
                 if metaslot_name == 'range':
                     v = self.schema.default_range
             if v is not None:
-                setattr(islot, metaslot_name, v)
+                setattr(induced_slot, metaslot_name, v)
         if slot.inlined_as_list:
             slot.inlined = True
         if mangle_name:
             mangled_name = f'{camelcase(class_name)}__{underscore(slot_name)}'
-            islot.name = mangled_name
-        if not islot.alias:
-            islot.alias = underscore(slot_name)
+            induced_slot.name = mangled_name
+        if not induced_slot.alias:
+            induced_slot.alias = underscore(slot_name)
         for c in self.all_classes().values():
-            if islot.name in c.slots or islot.name in c.attributes:
-                islot.domain_of.append(c.name)
-        return deepcopy(islot)
+            if induced_slot.name in c.slots or induced_slot.name in c.attributes:
+                induced_slot.domain_of.append(c.name)
+        return deepcopy(induced_slot)
 
     @lru_cache()
     def _metaslots_for_slot(self):
@@ -1307,7 +1349,7 @@ class SchemaView(object):
         :param include_induced: supplement all direct slots with induced slots, defaults to False
         :return: list of slots, either direct, or both direct and induced
         """
-        slots_list = [] # list of all direct or induced slots
+        slots_list = []  # list of all direct or induced slots
 
         for c_name, c in self.all_classes().items():
             # check if slot is direct specification on class
@@ -1338,20 +1380,23 @@ class SchemaView(object):
                         enum_slots.append(slot_definition)
         return enum_slots
 
-    def is_type_percent_encoded(self, type: TypeDefinitionName) -> bool:
+    def is_slot_percent_encoded(self, slot: SlotDefinitionName) -> bool:
         """
-        True if type is has a percent_encoded annotation.
+        True if slot or its range is has a percent_encoded annotation.
 
         This is true for type fields that are the range of identifier columns,
         where the identifier is not guaranteed to be a valid URI or CURIE
 
-        :param type:
+        :param slot:
         :return:
         """
-        id_slot_ranges = self.type_ancestors(type)
-        for t in id_slot_ranges:
-            anns = self.get_type(t).annotations
-            return "percent_encoded" in anns
+        if "percent_encoded" in slot.annotations:
+            return True
+        if slot.range in self.all_types():
+            id_slot_ranges = self.type_ancestors(slot.range)
+            for t in id_slot_ranges:
+                anns = self.get_type(t).annotations
+                return "percent_encoded" in anns
 
     @lru_cache()
     def usage_index(self) -> Dict[ElementName, List[SchemaUsage]]:
@@ -1470,30 +1515,38 @@ class SchemaView(object):
         del self.schema.subsetes[subset_name]
         self.set_modified()
 
-    #def rename(self, old_name: str, new_name: str):
+    # def rename(self, old_name: str, new_name: str):
     #   todo: add to runtime
 
-    def merge_schema(self, schema: SchemaDefinition) -> None:
+    def merge_schema(self, schema: SchemaDefinition, clobber=False) -> None:
         """
-        merges another schema into this one
+        merges another schema into this one.
+
+        If the other schema has an element with the same name as an element in this schema,
+        then this element is NOT copied.
+
         :param schema: schema to be merged
+        :param clobber: if True, then overwrite existing elements
         """
         dest = self.schema
         for k, v in schema.prefixes.items():
-            if k not in dest.prefixes:
+            if clobber or k not in dest.prefixes:
                 dest.prefixes[k] = copy(v)
         for k, v in schema.classes.items():
-            if k not in dest.classes:
+            if clobber or k not in dest.classes:
                 dest.classes[k] = copy(v)
         for k, v in schema.slots.items():
-            if k not in dest.slots:
+            if clobber or k not in dest.slots:
                 dest.slots[k] = copy(v)
         for k, v in schema.types.items():
-            if k not in dest.types:
+            if clobber or k not in dest.types:
                 dest.types[k] = copy(v)
         for k, v in schema.enums.items():
-            if k not in dest.types:
+            if clobber or k not in dest.enums:
                 dest.enums[k] = copy(v)
+        for k, v in schema.subsets.items():
+            if clobber or k not in dest.subsets:
+                dest.subsets[k] = copy(v)
         self.set_modified()
 
     def merge_imports(self):
@@ -1538,7 +1591,7 @@ class SchemaView(object):
             if class_definition.slot_usage:
                 for slot_definition in class_definition.slot_usage.values():
                     materialize_pattern_into_slot_definition(slot_definition)
-        
+
             if class_definition.attributes:
                 for slot_definition in class_definition.attributes.values():
                     materialize_pattern_into_slot_definition(slot_definition)
